//
// Copyright 2006 Mentor Graphics Corporation
//
// All Rights Reserved.
//
// THIS WORK CONTAINS TRADE SECRET AND PROPRIETARY INFORMATION WHICH IS THE PROPERTY
// OF MENTOR GRAPHICS CORPORATION OR ITS LICENSORS AND IS SUBJECT TO LICENSE TERMS.
//

`timescale 1 ns / 1 ns

module cache(paddr, pdata, prw, pstrb, prdy, p0);

    input  paddr, prw, pstrb;
    output prdy;
    inout  pdata;

    /**************** modport ̃CX^X ****************/
    pmbus.bus_p p0;

    `define addr_size  8
    `define set_size   5
    `define word_size  16

    logic [`word_size-1:0] mdata_r, pdata_r;
    logic [`addr_size-1:0] maddr_r;
    logic                  mrw_r, mstrb_r, prdy_r;

    logic [`word_size-1:0] ob_data;

    logic sysacc_r;
    wire #(5) sysacc = sysacc_r;

    logic aa;
    wire mstrb_x = mstrb_r && aa;

    wire [`addr_size-1:0]  paddr;
    assign #(5) p0.ADDR = maddr_r;
    assign #(5) p0.DATA = mdata_r;
    assign #(5) p0.RW = mrw_r;
    assign #(5) p0.STRB = mstrb_x;

    wire [`word_size-1:0] #(5) pdata = pdata_r;
    wire                  #(5) prdy = prdy_r;

    logic [3:0] oen, wen;
    logic [3:0] hit;

    /**************** JobW` ****************/

    // `include "cover.inc"

    /**************** Cache sets ****************/
    cache_set s0(paddr, pdata, hit[0], oen[0], wen[0]);
    cache_set s1(paddr, pdata, hit[1], oen[1], wen[1]);
    cache_set s2(paddr, pdata, hit[2], oen[2], wen[2]);
    cache_set s3(paddr, pdata, hit[3], oen[3], wen[3]);

    initial begin
        maddr_r = 0;
        mdata_r = 'bz;
        pdata_r = 'bz;
        mrw_r = 0;
        mstrb_r = 1;
        prdy_r = 1;
        oen = 4'b1111;
        wen = 4'b1111;
        aa = 1;
    end

    /**************** Local LRU memory ****************/

    logic [2:0] lru_mem [0:(1 << `set_size) - 1];

    integer i;
    initial for (i = 0; i < (1 << `set_size); i=i+1) lru_mem[i] = 0;

    function integer hash;
        input [`addr_size-1:0] a;
        hash = a[`set_size - 1:0];
    endfunction

    task update_lru;
        input [`addr_size-1:0] addr;
        input [3:0] hit;
        logic [2:0] lru;
        begin
            lru = lru_mem[hash(addr)];
            lru[2] = ((hit & 4'b1100) != 0);
            if (lru[2]) lru[1] = hit[3];
            else        lru[0] = hit[1];
            lru_mem[hash(addr)] = lru;
        end
    endtask

    function [3:0] pick_set;
        input [`addr_size-1:0] addr;
        integer setnum;
        begin
            casez (lru_mem[hash(addr)])
                3'b1?1 : setnum = 0;
                3'b1?0 : setnum = 1;
                3'b01? : setnum = 2;
                3'b00? : setnum = 3;
                default: setnum = 0;
            endcase
            if (prw == 1) if (top.verbose)
                $display("%t: Read miss, picking set %0d", $time, setnum);
            else if (top.verbose)
                $display("%t: Write miss, picking set %0d", $time, setnum);
            pick_set = 4'b0001 << setnum;
        end
    endfunction

    /**************** Cache control ****************/

    function [3:0] get_hit;
        input [3:0] hit;
        integer setnum;
        begin
            casez (hit)
                4'b???1 : setnum = 0;
                4'b??1? : setnum = 1;
                4'b?1?? : setnum = 2;
                4'b1??? : setnum = 3;
            endcase
            if (prw == 1) if (top.verbose)
                $display("%t: Read hit to set %0d", $time, setnum);
            else if (top.verbose)
                $display("%t: Write hit to set %0d", $time, setnum);
            get_hit = 4'b0001 << setnum;
        end
    endfunction


    /**************** Memory Access ****************/

    always @(posedge p0.CLK) if (sysacc) begin
        maddr_r = paddr;
        mstrb_r = 0;
        if (prw == 1) begin
            mrw_r = 1;
            @(posedge p0.CLK) mstrb_r = 1;
            assign prdy_r = p0.RDY;
            assign pdata_r = p0.DATA;
            @(posedge p0.CLK) while (p0.RDY != 0) @(posedge p0.CLK) ;
            deassign prdy_r;  prdy_r = 1;
            deassign pdata_r; pdata_r = 'bz;
        end else begin
            maddr_r = paddr;
            mrw_r = 0;
            mstrb_r = 0;
            @(posedge p0.CLK) mstrb_r = 1;
            assign prdy_r = p0.RDY;
            assign mdata_r = pdata;
            @(posedge p0.CLK) while (p0.RDY != 0) @(posedge p0.CLK) ;
            deassign prdy_r;  prdy_r = 1;
            deassign mdata_r; mdata_r = 'bz;
        end
        sysacc_r = 0;
    end

    /**************** Hit/Miss Operation ****************/

    logic [3:0] setsel;

    always @(posedge p0.CLK) if (pstrb == 0) begin
        if ((prw == 1) && hit) begin
            // Read Hit..
            setsel = get_hit(hit);
            oen = ~setsel;
            prdy_r = 0;
            @(posedge p0.CLK) prdy_r = 1;
            oen = 4'b1111;
        end else begin
            // Read Miss or Write Hit..
            if (hit)
                setsel = get_hit(hit);
            else
                setsel = pick_set(paddr);
            wen = ~setsel;
            sysacc_r = 1;
            @(posedge p0.CLK);
            wen = 4'b1111;
        end
        update_lru(paddr, setsel);
    end

endmodule

module cache_set(addr, data, hit, oen, wen);
  input addr, oen, wen;
  inout data;
  output hit;
  
  wire  [`addr_size-1:0] addr;
  logic [`word_size-1:0] data_r;
  logic hit_r;
  
  wire [`word_size-1:0] data = data_r;
  wire hit = hit_r;

  `define size (1 << `set_size)
  `define dly 5
  logic [`word_size-1:0] data_out;

  // ---------- Local tag and data memories -----------
  logic [`word_size-1:0] data_mem[0:(1 << `set_size)-1];
  logic [`addr_size-1:`set_size] atag_mem[0:(1 << `set_size)-1];
  logic [0:(1 << `set_size)-1] valid_mem;
  
  always @(data_out or oen)
    data_r <= #(5) oen ? `word_size'bz : data_out;

  function integer hash;
     input [`addr_size-1:0] a;
     hash = a[`set_size - 1:0];
  endfunction

  task lookup_cache;
    input [`addr_size-1:0] a;
    integer i;
    logic found;
  begin
    i = hash(a);
    found = valid_mem[i] && (a[`addr_size-1:`set_size] == atag_mem[i]);
    if (found) 
      hit_r <= #5 1'b1;
    else
      hit_r <= #5 1'b0;
  end
  endtask

  task update_cache;
    input [`addr_size-1:0] a;
    input [`word_size-1:0] d;
    integer i;
  begin
    i = hash(a);
    data_mem[i] = d;
    atag_mem[i] = a[`addr_size-1:`set_size];
    valid_mem[i] = 1'b1;
  end
  endtask
  
  integer i;
  initial begin
    for (i=0; i<`size; i=i+1)
      valid_mem[i] = 0;
  end

  always @(negedge(wen) or addr)
  begin
    lookup_cache(addr);
    data_out <= data_mem[hash(addr)];
  end
  
  always @(posedge(wen))
  begin
    update_cache(addr, data);
    lookup_cache(addr);
    data_out <= data_mem[hash(addr)];
  end
endmodule